/*
 *  Copyright (c) xlightweb.org, 2008. All rights reserved.
 *
 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Lesser General Public
 *  License as published by the Free Software Foundation; either
 *  version 2.1 of the License, or (at your option) any later version.
 *
 *  This library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Lesser General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public
 *  License along with this library; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Please refer to the LGPL license at: http://www.gnu.org/copyleft/lesser.txt
 * The latest copy of this software may be found on http://www.xlightweb.org/
 */
package org.xlightweb;

import java.io.IOException;
import java.util.logging.Level;
import java.util.logging.Logger;

import org.xsocket.connection.INonBlockingConnection;




/**
 * protcol handler implmentation base 
 *  
 * @author grro@xlightweb.org
 */
abstract class AbstractHttpProtocolHandler {
    
    private static final Logger LOG = Logger.getLogger(AbstractHttpProtocolHandler.class.getName());
    
    
	
	static final int BODY_TYPE_EMTPY = 0;
	static final int BODY_TYPE_BOUND = 1;
	static final int BODY_TYPE_CHUNKED = 2;
	static final int BODY_TYPE_SIMPLE = 3;
	static final int BODY_TYPE_MULTIPART_BYTERANGE = 4;
	
	static final int INIT = 0;
	static final int RECEIVING_HEADER = 5;
	static final int RECEIVING_BODY = 10;
	
	private int state = INIT;
	private IMessageHeaderParser headerParser = null;
	private AbstractBodyParser bodyParser = null;

	
	/**
	 * returns the state
	 * @return  the state
	 */
	protected final int getState() {
		return state;
	}
	
	/**
	 * sets the state
	 * @param state the state
	 */
	protected final void setState(int state) {
		this.state = state;
	}
	
	
	/**
	 * returns the body parser
	 * @return the body parser
	 */
	protected final AbstractBodyParser getBodyParser() {
		return bodyParser;
	}

	/**
	 * sets the body parser
	 * @param bodyParser  the body parser
	 */
	protected final void setBodyParser(AbstractBodyParser bodyParser) {
		this.bodyParser = bodyParser;
	}
	
	
	/**
	 * returns the header parser
	 * @return the header parser
	 */
	protected IMessageHeaderParser getHeaderParser() {
		return headerParser;
	}

	
	/**
	 * sets the header parser
	 * @param headerParser the header parser
	 */
	protected void setHeaderParser(IMessageHeaderParser headerParser) {
		this.headerParser = headerParser;
	}

	
	/**
	 * parses the body 
	 * 
	 * @param rawData    the raw network data
	 * @throws IOException if an exception occurs  
	 */
	protected final void parserBody(ComposedByteBuffer rawData) throws IOException {
		boolean isComplete = getBodyParser().parse(rawData);
		if (isComplete) {
			reset();
		}
	}

	
	/**
	 * reset the state
	 */
	protected final void reset() {
		headerParser.recycle();
		headerParser = null;
		bodyParser = null;
		state = INIT;
	}
	
	
	/**
	 * call back if the underlying connection is disconnected
	 */
	public final void onDisconnect() {
		if (bodyParser != null) {
			bodyParser.onDisconnect();
		}
	}
	
	/**
	 * call back if data is received
	 * 
	 * @param httpConnection        the http connection 
	 * @param underlyingConnection  the underyling tcp connection
	 * @param rawData               the raw network data recevie buffer
	 * @throws BadMessageException  if the parsed message is invalid 
	 * @throws IOException if an exception occurs
	 */
	protected abstract void onData(AbstractHttpConnection httpConnection, INonBlockingConnection underlyingConnection, ComposedByteBuffer rawData) throws BadMessageException, IOException;
	
	
	/**
     * call back if a exception is occured
     * 
     * @param  ioe  the exception
     */
    protected void onException(IOException ioe, ComposedByteBuffer rawData) {
        
        switch (getState()) {
        
            case INIT:
                if (LOG.isLoggable(Level.FINE)) {
                    LOG.fine("Error occured while waiting for header " + ioe.toString());
                }
                return;
        
            case RECEIVING_HEADER:
                
                IMessageHeaderParser headerParser = getHeaderParser();
                if (headerParser != null) {
                    headerParser.onException(ioe, rawData);
                }
                break;
    
                
            default:
                AbstractBodyParser bodyParser = getBodyParser();
                if (bodyParser != null) {
                    bodyParser.onException(ioe, rawData);
                }
        }
    }
}

 