/*
 *  Copyright (c) xlightweb.org, 2008. All rights reserved.
 *
 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Lesser General Public
 *  License as published by the Free Software Foundation; either
 *  version 2.1 of the License, or (at your option) any later version.
 *
 *  This library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Lesser General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public
 *  License along with this library; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Please refer to the LGPL license at: http://www.gnu.org/copyleft/lesser.txt
 * The latest copy of this software may be found on http://www.xlightweb.org/
 */
package org.xlightweb;

import java.io.Flushable;
import java.io.IOException;
import java.nio.channels.ClosedChannelException;
import java.util.ArrayList;
import java.util.concurrent.atomic.AtomicBoolean;
import java.util.logging.Level;
import java.util.logging.Logger;

import org.xlightweb.AbstractHttpConnection.IMultimodeExecutor;
import org.xsocket.IDataSink;
import org.xsocket.connection.AbstractNonBlockingStream;
import org.xsocket.connection.IConnection.FlushMode;




/**
 * I/O resource capable of receiving the body data.
 * 
 * @author grro@xlightweb.org
 *
 */
public final class BodyDataSink extends AbstractNonBlockingStream implements IDataSink, Flushable {
	
	private static final Logger LOG = Logger.getLogger(BodyDataSink.class.getName());
	
	
	private final ArrayList<IBodyCloseListener> closeListeners = new ArrayList<IBodyCloseListener>();
	
	private final IMultimodeExecutor executor;
	private final AbstractHttpConnection httpConnection;

	private IBodyWriter bodySerializer;
	private AtomicBoolean isOpen = new AtomicBoolean(true);
	
	

	/**
	 * constructor 
	 * 
	 * @param httpConnection     the underlying http connection
	 * @param executor           the executor
	 * @param bodySerializer     the body serializer
	 * @param characterEncoding  the character encoding
	 * @throws IOException if an exception occurs
	 */
	BodyDataSink(AbstractHttpConnection httpConnection, IMultimodeExecutor executor, IBodyWriter bodySerializer, String characterEncoding) throws IOException {
	
		this.httpConnection = httpConnection;
		this.executor = executor;
		this.bodySerializer = bodySerializer;
		
		if (httpConnection != null) {
			this.httpConnection.setBodyDataSink(this);
		}
		
		setFlushmode(FlushMode.SYNC);
		setEncoding(characterEncoding);
	}

	
	
	/**
	 * returns the body writer
	 * @return the body writer
	 */
	IBodyWriter getBodyWriter() {
		return bodySerializer;
	}
	
	
	/**
	 * sets the body writer
	 * @param bodyWriter  the body writer
	 */
	void setBodyWriter(IBodyWriter bodyWriter) {
		this.bodySerializer = bodyWriter;
	}

	
	/**
	 * adds a close listener 
	 * @param closeListener the close listener
	 */
	void addCloseListener(IBodyCloseListener closeListener) {
		synchronized (closeListeners) {
			closeListeners.add(closeListener);
		}
	}

	
	/**
	 * remove a close listener
	 * 
	 * @param closeListener a close listener
	 * @return true, if the listener is removed
	 */
	boolean removeCloseListener(IBodyCloseListener closeListener) {
		synchronized (closeListeners) {
			return closeListeners.remove(closeListener);
		}
	}
	
	
	/**
	 * {@inheritDoc}
	 */
	public void flush() {
		
		removeWriteMark();
		
		try {
			bodySerializer.flush(drainWriteQueue(), getFlushmode());
		} catch (IOException ioe) {
			throw new RuntimeException(ioe.toString());
		}
	}
	
	
	
	/**
	 * {@inheritDoc}
	 */
	@Override
	protected void onWriteDataInserted() throws IOException, ClosedChannelException {
		
		if (isAutoflush()) {
			try {
				flush();
			} catch (Exception e) {
				throw new IOException(e.toString());
			}
		}
	}

	
	
	/**
	 * {@inheritDoc}
	 */
	public void close() throws IOException {

		isOpen.set(false);
			
		if (httpConnection != null) {
			httpConnection.removeBodyDataSink(this);
		}
			
		try {
			flush();
		} catch (Exception e) {
			throw new IOException(e.toString());
				
		} finally {
			if (LOG.isLoggable(Level.FINE)) {
				LOG.fine("closing body serializer " + bodySerializer);
			}
			bodySerializer.close();	
			callCloseListener();
		}
	}
	
	
	
	/**
	 * returns true if the data sink is open
	 * 
	 * @return true if the data sink is open
	 */
	public boolean isOpen() {
		return isOpen.get();
	}
	

	
	/**
	 * {@inheritDoc}
	 */
	@Override
	protected boolean isDataWriteable() {
		return isOpen.get();
	}

	
	/**
	 * {@inheritDoc}
	 */
	@Override
	protected boolean isMoreInputDataExpected() {
		return isOpen.get();
	}
	
	
	/**
	 * call back if the underlying connection is closed
	 */
	void onUnderlyingHttpConnectionClosed() {
		if (isOpen.get()) {
			
			if (LOG.isLoggable(Level.FINE)) {
				LOG.info("underlying connection is closed. clsoing data source");
			}
			isOpen.set(false);
			
			bodySerializer.destroy();
			callCloseListener();
		}
	}
	
	
	
	@SuppressWarnings("unchecked")
	private void callCloseListener() {
		ArrayList<IBodyCloseListener> closeListenersCopy = null;
		synchronized (closeListeners) {
			closeListenersCopy = (ArrayList<IBodyCloseListener>) closeListeners.clone();
		}
		
		
		for (IBodyCloseListener bodyCloseListener : closeListenersCopy) {
			removeCloseListener(bodyCloseListener);
			callCloseListener(bodyCloseListener);
		}
	}
	
	
	
	private void callCloseListener(IBodyCloseListener listener) {
		
		Runnable task = new CloseListenerCaller(listener);
		
		if (HttpUtils.isBodyCloseListenerMutlithreaded(listener)) {
			executor.processMultithreaded(task);
		} else {
			executor.processNonthreaded(task);
		}
	}
	
	
	private static final class CloseListenerCaller implements Runnable {
	
		private IBodyCloseListener listener = null;
		
		public CloseListenerCaller(IBodyCloseListener listener) {
			this.listener = listener;
		}
		
		public void run() {
			try {
				listener.onClose();
			} catch (IOException ioe) {
				if (LOG.isLoggable(Level.FINE)) {
					LOG.fine("Error occured by calling close listener " + listener + " " + ioe.toString());
				}
			}			
		}
	}
	

	/**
	 * destroys this data sink  
	 */
	public void destroy() {
		isOpen.set(false);
		
		bodySerializer.destroy();
		callCloseListener();
	}
	
	
	/**
	 * {@inheritDoc}
	 */
	public String toString() {
		return printWriteBuffer(getEncoding());
	}	
}
