/*
 *  Copyright (c) xlightweb.org, 2008. All rights reserved.
 *
 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Lesser General Public
 *  License as published by the Free Software Foundation; either
 *  version 2.1 of the License, or (at your option) any later version.
 *
 *  This library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Lesser General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public
 *  License along with this library; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Please refer to the LGPL license at: http://www.gnu.org/copyleft/lesser.txt
 * The latest copy of this software may be found on http://www.xlightweb.org/
 */
package org.xlightweb;

import java.io.IOException;
import java.nio.BufferUnderflowException;
import java.nio.ByteBuffer;
import java.util.concurrent.atomic.AtomicBoolean;
import java.util.logging.Level;
import java.util.logging.Logger;

import org.xsocket.DataConverter;
import org.xsocket.Execution;



/**
 * Implementation base of a body forwarder. This class handles closing of he data source and exceptions by 
 * forwarding the body data. Example:
 * 
 * <pre>
 * 
 * class ForwardHandler implements IHttpRequestHandler {
 * 
 *   private HttpClient httpClient = ...
 *    
 *    
 *   // will be called if request header is received (Default-InvokeOn is HEADER_RECEIVED) 
 *   public void onRequest(IHttpExchange exchange) throws IOException {
 *      
 *      IHttpRequestHeader requestHeader = exchange.getRequest().getRequestHeader();
 *      NonBlockingBodyDataSource requestBodyChannel = exchange.getRequest().getNonBlockingBody();
 *      
 *       // forwards the request (header)
 *       BodyDataSink dataSink = httpClient.send(requestHeader);
 *       
 *       // defines the body forwarder
 *       BodyForwarder bodyForwarder = new BodyForwarder(requestBodyChannel, dataSink) {
 *       
 *          public void onData(NonBlockingBodyDataSource bodyDataSource, BodyDataSink bodyDataSink) throws BufferUnderflowException, IOException {
 *             int available = bodyDataSink.available();
 *             ByteBuffer[] data  = bodyDataSink.readByteBufferByLength(available);
 *             
 *             for (ByteBuffer byteBuffer : data) {
 *                ByteBuffer copy = byteBuffer.duplicate();
 *                // ...
 *             }
 *             
 *             bodyDataSink.write(data);
 *          }
 *       };
 *       
 *       // assign the body forwarder to the body data source 
 *       requestBodyChannel.setDataHandler(bodyForwarder);
 *    }
 * }
 * </pre>
 * 
 * 
 * @author grro@xlightweb.org
 */
public class BodyForwarder implements IBodyDataHandler {
	
	private static final Logger LOG = Logger.getLogger(BodyForwarder.class.getName());
	
	public static final int DEFAULT_EXECUTION_MODE = IBodyDataHandler.DEFAULT_EXECUTION_MODE;
	
	private final ForwardTask forwardTask = new ForwardTask();
	private final DataSinkCloseHandler dataSinkCloseHandler = new DataSinkCloseHandler();
	
	private final NonBlockingBodyDataSource bodyDataSource;
	private final BodyDataSink bodyDataSink;
	
	private AtomicBoolean isComplete = new AtomicBoolean(false); 
	
	
	/**
	 * constructor
	 * 
	 * @param bodyDataSource   the body data source 
	 * @param bodyDataSink     the body data sink
	 */
	public BodyForwarder(final NonBlockingBodyDataSource bodyDataSource, BodyDataSink bodyDataSink) {
		this.bodyDataSource = bodyDataSource;
		this.bodyDataSink = bodyDataSink;
		
		bodyDataSink.addCloseListener(dataSinkCloseHandler);
	}
	
	
	/**
	 * {@inheritDoc}
	 */
	@Execution(Execution.NONTHREADED)
	public final boolean onData(final NonBlockingBodyDataSource bodyDataSource) throws BufferUnderflowException {
		
		//	TODO remove this
		if (this.bodyDataSource != bodyDataSource) {
			return true; 
		}
		
		
		try {
			int available = bodyDataSource.available();
			
			// data available
			if (available > 0) {
				if (HttpUtils.isMutlithreaded(this)) {
					bodyDataSource.processMultithreaded(forwardTask);
				} else {
					bodyDataSource.processNonthreaded(forwardTask);
				}
				
			// end of stream reached
			} else if (available == -1) {
				handleEndOfSourceStream();  
			}

		} catch (Exception e) {
			if (LOG.isLoggable(Level.FINE)) {
				LOG.fine("[" + bodyDataSource.getId() + "] error occured by writing data sink " + bodyDataSource + " " + e.toString() + " destroying data source and data sink");
			}
			
			handleException(e);
		}
		
		
		return true;
	}
	
	
	
	private void handleEndOfSourceStream() {
		isComplete.set(true);
		
		detachBodyDataSource();   // detach the link   dataSource -> body forwarder 
		callOnComplete();
		closeDataSink();          // detach the link   body forwarder -> bodyDataSink	
	}
	
	
	
	private void detachBodyDataSource() {
		bodyDataSource.setDataHandler(null);
	}
	
	
	
	private void handleException(final Exception e) {
		destroy();
		
		Runnable task = new Runnable() {
			public void run() {
				onException(e);
			}
		};
		bodyDataSource.processMultithreaded(task);
	}
	
	
	private void destroy() {
		bodyDataSink.destroy();      // detach the link   dataSource -> body forwarder 
		bodyDataSource.destroy();    // detach the link   dataSource -> body forwarder 

		isComplete.set(true);
	}
	
	

	private void closeDataSink() {
		isComplete.set(true);
		
		if (LOG.isLoggable(Level.FINE)) {
			LOG.fine("[" + bodyDataSource.getId() + "] data source " + bodyDataSource.getClass().getSimpleName() + " is closed. Closing data sink " + bodyDataSink.getClass().getSimpleName());
		}
		try {
			bodyDataSink.close();
		} catch (Exception e) {
			handleException(e);
		}
	}
	
	
	
	
	/**
	 * call back method which will be called if the body data source contains a least one byte
	 * 
	 * @param bodyDataSource   the body data source 
	 * @param bodyDataSink     the body data sink
	 * @throws BufferUnderflowException if not enough data is available 
	 * @throws IOException if an exception occurs 
	 */
	public void onData(NonBlockingBodyDataSource bodyDataSource, BodyDataSink bodyDataSink) throws BufferUnderflowException, IOException {
		ByteBuffer[] data = bodyDataSource.readByteBufferByLength(bodyDataSource.available());
		
		bodyDataSink.write(data);
		bodyDataSink.flush();
	}
	
	
	private void callOnComplete() {
		
		try {
			onComplete();
		} catch (Exception e) {
			if (LOG.isLoggable(Level.FINE)) {
				LOG.fine("error occured by calling onComplete "  + e.toString());
			}
			throw new RuntimeException(e.toString());
		}
	}
		
	/**
	 * call back which will be executed, if the body data source's end of data is reached 
	 */
	public void onComplete() {
		
	}

	
	/**
	 * call back which will be executed, if an exception is occurred
	 * 
	 * @param e the exception
	 */
	public void onException(Exception e) {
		
	}
	
	
	
	
	private final class DataSinkCloseHandler implements IBodyCloseListener {
		
		public void onClose() throws IOException {
			if (!isComplete.get()) {
				destroy();
			}
		}
	}


	
	private final class ForwardTask implements Runnable {
			
		public void run() {
			try {
				while (bodyDataSource.available() > 0) {
					int version = bodyDataSource.getReadBufferVersion();

					onData(bodyDataSource, bodyDataSink);
						
					if (version == bodyDataSource.getReadBufferVersion()) {
						break;
					}
				}
				
				if (bodyDataSource.available() == -1) {
					handleEndOfSourceStream();
				}
			} catch (IOException e) {
				if (LOG.isLoggable(Level.FINE)) {
					LOG.fine("error occured by forwarding data " + DataConverter.toString(e));
				}
				handleException(e);
			}
		}		
	}
}
