/*
 *  Copyright (c) xlightweb.org, 2008. All rights reserved.
 *
 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Lesser General Public
 *  License as published by the Free Software Foundation; either
 *  version 2.1 of the License, or (at your option) any later version.
 *
 *  This library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Lesser General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public
 *  License along with this library; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Please refer to the LGPL license at: http://www.gnu.org/copyleft/lesser.txt
 * The latest copy of this software may be found on http://www.xlightweb.org/
 */
package org.xlightweb;

import java.io.IOException;
import java.nio.ByteBuffer;
import java.util.logging.Level;
import java.util.logging.Logger;

import org.xsocket.DataConverter;
import org.xsocket.connection.IConnection.FlushMode;




/**
 * chunked message body writer  
 * 
 * @author grro@xlightweb.org
 */
final class FullMessageChunkedWriter extends AbstractBodyWriter {
	
	private static final Logger LOG = Logger.getLogger(FullMessageChunkedWriter.class.getName());
	
	private boolean isHeaderWritten = false;

	private int writtenData;
	
	private final IHttpHeader header;
	private boolean isOpen = true;

	
	/**
	 * constructor
	 * 
	 * @param httpConnection   the http connection
	 * @param header           the message header
	 * @param encoding         the encoding to use
	 */
	FullMessageChunkedWriter(AbstractHttpConnection httpConnection, IHttpHeader header) throws IOException {
		super(httpConnection);
		
		this.header = header;
	}
	
		
	
	private void writeChunk(ByteBuffer[] bodyData, FlushMode flushMode) throws IOException {
		
		AbstractHttpConnection con = getConnection();
		if (con != null) {
			con.setFlushmode(flushMode);
			int written = 0;
			
			
			// write header (if not already written)
			if (!isHeaderWritten) {
				isHeaderWritten = true;
				
				if (LOG.isLoggable(Level.FINE)) {
					LOG.fine("[" + con.getId() + "] sending header ");
				}
				
				written += con.write(header.toString() + "\r\n");			
			}
	
			
			
			// write body (if content size larger than 0)
			int length = 0;
			
			if (bodyData != null) {
				for (ByteBuffer buffer : bodyData) {
					length += buffer.remaining();
				}
			}
			
			if (length > 0) {
				if (LOG.isLoggable(Level.FINE)) {
					LOG.fine("[" + con.getId() + "] writing chunk (size=" + length + ")");
				}
				
				// write length field
				written += con.write(Integer.toHexString(length) + "\r\n");
				
			
				
				// write data 
				written += con.write(bodyData);
				written += con.write("\r\n");
				
				writtenData += length;
			}
			
			if (written > 0) {
				con.flush();
			}
		}
	}

	
	/**
	 * {@inheritDoc}
	 */
	public void flush(ByteBuffer[] bodyData, FlushMode flushMode) throws IOException {
		
		try {
			
			writeChunk(bodyData, flushMode);
			
		} catch (IOException ioe) {
			if (getConnection() != null) {
				if (LOG.isLoggable(Level.FINE)) {
					LOG.fine("error occured by flushing chunked data sink. Destroying connection. reason " + DataConverter.toString(ioe));
				}
			
				destroy();
				throw ioe;
			}	
		}
	}
	

	/**
	 * {@inheritDoc}
	 */
	public void close() throws IOException {

		if (isOpen) {
			isOpen = false;
			
			AbstractHttpConnection con = getConnection();
			if (con == null) {
				return;
			}
		
			try {
				
				if (LOG.isLoggable(Level.FINE)) {
					LOG.fine("[" + con.getId()+ "] closing chunked body by writing termination chunk (body size=" + writtenData + ")");
				}

				// write termination chunk
				con.write("0" + "\r\n\r\n");
				con.flush();
			
			
			} catch (IOException ioe) {
				if (LOG.isLoggable(Level.FINE)) {
					LOG.fine("error occured by closing chunked data sink. Destroying connection. reason " + DataConverter.toString(ioe));
				}
				
				con.destroy();
			}
			
			super.close();
		}
	}
}
