/*
 *  Copyright (c) xlightweb.org, 2008. All rights reserved.
 *
 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Lesser General Public
 *  License as published by the Free Software Foundation; either
 *  version 2.1 of the License, or (at your option) any later version.
 *
 *  This library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Lesser General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public
 *  License along with this library; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Please refer to the LGPL license at: http://www.gnu.org/copyleft/lesser.txt
 * The latest copy of this software may be found on http://www.xlightweb.org/
 */
package org.xlightweb;

import java.io.IOException;
import java.net.SocketTimeoutException;

import org.xsocket.Execution;




/**
 * A response handler implementation which supports a future behavior. Typically this 
 * response handler will be used, if the send method is required to send the request,
 * but the response should be read in a blocking behavior. Example:  
 * 
 * <pre>
 *   FutureResponseHandler future = new FutureResponseHandler();
 *    
 *   HttpRequestHeader header = new HttpRequestHeader("POST", url, "application/octet-stream");
 *   
 *   BodyDataSink bodyDataSink = httpClient.send(header, future);
 *   bodyDataSink.transferFrom(source);
 *   bodyDataSink.close();
 *   
 *   IHttpResponse response = future.getResponse(); // blocks until the response header is received
 *   if (response.getStatus() != 200) {
 *      throw new IOException("got status " + response.getStatus());
 *   }
 * </pre>
 *    
 *  
 * @author grro@xlightweb.org
 */
@Execution(Execution.NONTHREADED)
public final class FutureResponseHandler implements IHttpResponseHandler, IHttpSocketTimeoutHandler {


	private final Object readLock = new Object();
	
	private IHttpResponse response;
	private IOException ioException;
	private SocketTimeoutException stException;
	

	/**
	 * {@inheritDoc}
	 */
	@InvokeOn(InvokeOn.HEADER_RECEIVED)
	public void onResponse(IHttpResponse response) throws IOException {
		
		synchronized (readLock) {
			this.response = response;
			readLock.notifyAll();
		}
	}


	/**
	 * {@inheritDoc}
	 */
	public void onException(IOException ioe) {
		
		synchronized (readLock) {
			this.ioException = ioe;
			readLock.notifyAll();
		}
	}

	
	/**
	 * {@inheritDoc}
	 */
	public void onException(SocketTimeoutException stoe) {
		
		synchronized (readLock) {
			this.stException = stoe;
			readLock.notifyAll();
		}		
	}
	
	
	/**
	 * blocking call to retrieve the response
	 *  
	 * @return the response 
	 * @throws IOException             if an ioe exception occurs
	 * @throws SocketTimeoutException  if an socket timeout exception occurs
	 */
	public IHttpResponse getResponse() throws IOException, SocketTimeoutException {	

		synchronized (readLock) {

			while (true) {
				if (stException != null) {
					throw stException;
				}
				
				if (ioException != null) {
					throw ioException;
				}
				
				if (response != null) {
					return response;
					
				}
				
				try {
					readLock.wait();
				} catch (InterruptedException ignore) { }
			}
		}
	}
}
