/*
 *  Copyright (c) xlightweb.org, 2008. All rights reserved.
 *
 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Lesser General Public
 *  License as published by the Free Software Foundation; either
 *  version 2.1 of the License, or (at your option) any later version.
 *
 *  This library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Lesser General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public
 *  License along with this library; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Please refer to the LGPL license at: http://www.gnu.org/copyleft/lesser.txt
 * The latest copy of this software may be found on http://www.xlightweb.org/
 */
package org.xlightweb;


import java.io.IOException;
import java.util.logging.Level;
import java.util.logging.Logger;

import org.xsocket.connection.INonBlockingConnection;

import static org.xlightweb.HttpUtils.*;


/**
 * http request header parser
 * 
 * @author grro@xlightweb.org
 */
final class HttpRequestHeaderParser implements IMessageHeaderParser {
	
    private static final Logger LOG = Logger.getLogger(HttpRequestHeaderParser.class.getName());
    
	
	private static ThreadLocal<HttpRequestHeaderParser> instanceThreadLocal = new ThreadLocal<HttpRequestHeaderParser>(); 
	

	private static final int MAX_HEADER_SIZE = 8192;
	
	
	private static final int STATE_INIT = 0;
	private static final int STATE_READING_REQUEST_LINE = 5;
	private static final int STATE_READING_HEADER_LINES = 9;
	private int state = STATE_INIT;


	private HttpRequestHeader requestHeader = null;

	private RequestLineParser requestLineParser;
	private HeaderlineParser headerlineParser;
	
	
	/**
	 * gets a new instance of the request header parser
	 * @return a new instance of the request header parser 
	 */
	static HttpRequestHeaderParser newInstance() {
		HttpRequestHeaderParser instance = instanceThreadLocal.get();
		if (instance == null) {
			instance = new HttpRequestHeaderParser();
		}
		
		instanceThreadLocal.set(null);
		
		return instance;
	}


	/**
	 * recycles a request header parser
	 * @param parser   the request header parser to recycle
	 */
	static void recycleInstance(HttpRequestHeaderParser parser) {
		instanceThreadLocal.set(parser);
	}
	
	
	
	/**
	 * {@inheritDoc}
	 */
	public void recycle() {
		HttpRequestHeaderParser.recycleInstance(this);
	}
	
	
	
	/**
	 * {@inheritDoc}
	 */
	public IHttpHeader parse(INonBlockingConnection connection, ComposedByteBuffer composedByteBuffer) throws BadMessageException, IOException {
		
		int size = 0;

		switch (state) {
		
			case STATE_INIT:
				requestHeader = new HttpRequestHeader(connection);
				requestLineParser = RequestLineParser.newInstance();
				state = STATE_READING_REQUEST_LINE;
				return parse(connection, composedByteBuffer);
				
				
			case STATE_READING_REQUEST_LINE:
				size = computeSizeToRead(composedByteBuffer);
				for (int i = 0; i < size; i++) {
					boolean isRead = requestLineParser.parse(connection, composedByteBuffer, requestHeader);
					if (isRead) {
						requestLineParser.recycle();
						requestLineParser = null;
						headerlineParser = HeaderlineParser.newInstance();
						
						state = STATE_READING_HEADER_LINES;
						return parse(connection, composedByteBuffer);
					}
				}
				break;
				
				
			case STATE_READING_HEADER_LINES:
				size = computeSizeToRead(composedByteBuffer);
				for (int i = 0; i < size; i++) {
					boolean isRead = headerlineParser.parse(composedByteBuffer, requestHeader);
					if (isRead) {
						headerlineParser.recycle();
						headerlineParser = null;

						HttpRequestHeader result = requestHeader;
						requestHeader = null;
						state = STATE_INIT;

						completeHeader(result);
						return result;
					}
				}
				break;
				
			default:
				break;
		}

		
		if (requestHeader.incCountParsedChars(size) >= MAX_HEADER_SIZE) {
			throw new BadMessageException("max header size exceeded");			
		}
	
		return null;
	}
	
	
	private void completeHeader(HttpRequestHeader header) {
		
		String host = header.getHeader("Host");
		int port = -1;
		
		if (host != null) {
			int idx = host.lastIndexOf(":");
			if (idx != -1) {
				port = Integer.parseInt(host.substring(idx + 1, host.length()).trim());
				host = host.substring(0, idx);
			}
			
			
			header.setServernameSilence(host);
			if (port != -1) {
				header.setServerportPortSilence(port);
			}
		} 
	}

	

	private int computeSizeToRead(ComposedByteBuffer composedByteBuffer) {
		int remaining = composedByteBuffer.remaining();
		
		
		// handle max header size
		int size = remaining;
		if (remaining > MAX_HEADER_SIZE - requestHeader.getCountParsedChars()) {
			size = MAX_HEADER_SIZE - requestHeader.getCountParsedChars();
		} 
		
		return size;
	}
	
	
	public void onException(IOException ioe, ComposedByteBuffer rawData) {
	    
	    if (LOG.isLoggable(Level.FINE)) {
	        LOG.info("error occured by receiving header. Reason: " + ioe.toString());
    	    StringBuilder sb = new StringBuilder();
    	    if (requestHeader != null) {
    	        sb.append(requestHeader.toString());
    	    } 

            LOG.info("received header lines: " +  sb.toString());
	    }
	}

		
	
	private static final class RequestLineParser {

		
		private static ThreadLocal<RequestLineParser> instanceThreadLocal = new ThreadLocal<RequestLineParser>();
		
		private int state = 0;
		
		
		private final StringBuilder methodBuilder = new StringBuilder(8);
		private final StringBuilder pathBuilder = new StringBuilder();
		private final StringBuilder queryBuilder = new StringBuilder();
		private final StringBuilder rawQueryBuilder = new StringBuilder();
		private final StringBuilder protocolBuilder = new StringBuilder(4);
		private final StringBuilder protocolVersionBuilder = new StringBuilder(3);
		
		private String paramName = null;
	
		
		static RequestLineParser newInstance() {
			RequestLineParser instance = instanceThreadLocal.get();
			if (instance == null) {
				instance = new RequestLineParser();
			}
			
			instanceThreadLocal.set(null);
			
			return instance;
		}
		
		static void recycleInstance(RequestLineParser parser) {
			instanceThreadLocal.set(parser);
		}
		
		
		public boolean parse(INonBlockingConnection connection, ComposedByteBuffer composedByteBuffer, HttpRequestHeader messageHeader) throws BadMessageException, IOException {

			byte b = composedByteBuffer.getByte();

			
			// filter CR:
			// RFC 2616 (19.3 Tolerant Applications)
			// ... The line terminator for message-header fields is the sequence CRLF.
			// However, we recommend that applications, when parsing such headers,
			// recognize a single LF as a line terminator and ignore the leading CR...

			
			switch (state) {
			
				 // tolerate leading CR, LF
				case 0:
					switch (b) {
						case CR:
							break;
							
						case LF:
							break;
							
						case SPACE:
							reset();
							throw new BadMessageException("bad request. request starts with SPACE");
							
						case HTAB:
							reset();
							throw new BadMessageException("bad request. request starts with HTAB");
	
						default:
							methodBuilder.append((char) b);
							state = 5;
							break;
					}
					break;

			
			    // until method read
				case 5:
					switch (b) {
						case SPACE:
							messageHeader.setMethodSilence(methodBuilder.toString());
							state = 10;
							break;
							
						case HTAB:
							state = 10;
							messageHeader.setMethodSilence(methodBuilder.toString());
							break;

						default:
							methodBuilder.append((char) b);
							break;
					}

					break;

					
				// wait until path starts
				case 10:
					switch (b) {
						case SPACE:
							break;
							
						case HTAB:
							break;
	
						default:
							pathBuilder.append((char) b);
							state = 11;
							break;
					}
					break;
					
				
				// wait for query params or end of path
				case 11:
					switch (b) {
						case SPACE:
							messageHeader.setPathSilence(pathBuilder.toString());
							state = 30;
							break;
							
						case HTAB:
							messageHeader.setPathSilence(pathBuilder.toString());
							state = 30;
							break;
							
						case '?':
							messageHeader.setPathSilence(pathBuilder.toString());
							state = 20;
							break;
							
						case CR:
							break;
							
						case LF:
							reset();
							throw new BadMessageException("simple request messages are not supported");

						default:
							pathBuilder.append((char) b);
							break;
					}
					break;
					
					
				// read first query char
				case 20:
					switch (b) {
						case SPACE:
							state = 30;
							break;
							
						case HTAB:
							state = 30;
							break;
							
						default:
							rawQueryBuilder.append((char) b);
							queryBuilder.append((char) b);
							state = 21;
							break;
					}
					break;

					
				// read query param name or end of query 
				case 21:
					switch (b) {
						case SPACE:
							messageHeader.setRawQueryStringSilence(rawQueryBuilder.toString());
							state = 30;
							break;
							
						case HTAB:
							messageHeader.setRawQueryStringSilence(rawQueryBuilder.toString());
							state = 30;
							break;
							
						case '=':
							rawQueryBuilder.append((char) b);
							paramName = queryBuilder.toString();
							queryBuilder.setLength(0);
							state = 22;
							break;
							
						case CR:
							break;
							
						case LF:
							messageHeader.setRawQueryStringSilence(rawQueryBuilder.toString());
							reset();
							return true;
							
						default:
							rawQueryBuilder.append((char) b);
							queryBuilder.append((char) b);
							break;
					}
					break;
					
					
				// read query param value or end of query 
				case 22:
					switch (b) {
						case SPACE:
							messageHeader.setRawQueryStringSilence(rawQueryBuilder.toString());
							messageHeader.addRawQueryParameterSilence(paramName, queryBuilder.toString());
							paramName = null;
							queryBuilder.setLength(0);
							state = 30;
							break;
							
						case HTAB:
							messageHeader.setRawQueryStringSilence(rawQueryBuilder.toString());
							messageHeader.addRawQueryParameterSilence(paramName, queryBuilder.toString());
							paramName = null;
							queryBuilder.setLength(0);
							state = 30;
							break;
							
						case '&':
							rawQueryBuilder.append((char) b);
							messageHeader.addRawQueryParameterSilence(paramName, queryBuilder.toString());
							paramName = null;
							queryBuilder.setLength(0);
						
							state = 21;
							break;
							
						case CR:
							break;
							
						case LF:
							messageHeader.setRawQueryStringSilence(rawQueryBuilder.toString());
							reset();
							return true;
							
						default:
							queryBuilder.append((char) b); 
							rawQueryBuilder.append((char) b);
							break;
					}
					break;
													
					
				// wait until protocol starts 
				case 30:
					switch (b) {
						case SPACE:
							break;
							
						case HTAB:
							break;
							
						case CR:
							break;
							
						case LF:
							reset();
							return true;
							
						default:
							protocolBuilder.append((char) b);
							state = 31;
							break;
					}
					break;
					
				// wait until '/' occurs
				case 31:
					switch (b) {
						case '/':
							String protocol = protocolBuilder.toString();
							messageHeader.setProtocolSchemeSilence(protocol);
							
							state = 32;
							break;
							
						case CR:
							break;
							
						case LF:
							protocol = protocolBuilder.toString();
							messageHeader.setProtocolVersionSilence(protocol);
							
							reset();
							return true;
							
						default:
							protocolBuilder.append((char) b);
							break;
					}
					break;
					
				// '/' occurred
				case 32:
					protocolVersionBuilder.append((char) b);
					state = 33;
					break;
					
				// wait until version ends
				case 33:
					switch (b) {
						case SPACE:
							messageHeader.setProtocolVersionSilence(protocolVersionBuilder.toString());
							state = 999;
							break;
							
						case HTAB:
							messageHeader.setProtocolVersionSilence(protocolVersionBuilder.toString());
							state = 999;
							break;
							
						case CR:
							break;
							
						case LF:
							messageHeader.setProtocolVersionSilence(protocolVersionBuilder.toString());
							reset();
							return true;

						default:
							protocolVersionBuilder.append((char) b);
							break;
					}
					break;

				default:
					break;
			}
			
			
			return false;
		}
		
		
		
				
		public void reset() {
			state = 0;
			
			methodBuilder.setLength(0);
			pathBuilder.setLength(0);
			queryBuilder.setLength(0);
			rawQueryBuilder.setLength(0);
			protocolBuilder.setLength(0);
			protocolVersionBuilder.setLength(0);
			
			paramName = null;
		}
		
		
		public void recycle() {
			RequestLineParser.recycleInstance(this);
		}
		
	}
}

 