/*
 *  Copyright (c) xlightweb.org, 2008. All rights reserved.
 *
 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Lesser General Public
 *  License as published by the Free Software Foundation; either
 *  version 2.1 of the License, or (at your option) any later version.
 *
 *  This library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Lesser General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public
 *  License along with this library; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Please refer to the LGPL license at: http://www.gnu.org/copyleft/lesser.txt
 * The latest copy of this software may be found on http://www.xlightweb.org/
 */
package org.xlightweb;


import java.net.URL;
import java.util.Enumeration;
import java.util.Set;






/**
 * Http Request Header
 * 
 * @author grro@xlightweb.org
 */
public interface IHttpRequestHeader extends IHttpHeader {
	 
	
    
	/**
	 * Returns the fully qualified name of the client 
	 * or the last proxy that sent the request. If the 
	 * engine cannot or chooses not to resolve the hostname
	 * (to improve performance), this method returns the 
	 * dotted-string form of the IP address.
	 * 
	 * @return a String containing the fully qualified name of the client
	 */
	String getRemoteHost();
	
	
	/**
	 * Returns the Internet Protocol (IP) source 
	 * port of the client or last proxy that sent the request.
	 * 
	 * @return an integer specifying the port number
	 */
	int getRemotePort();
	

	/**
	 * Returns the Internet Protocol (IP) address of the client or last proxy that
	 * sent the request.
	 * 
	 * @return a String containing the IP address of the client that sent the request
	 */
	String getRemoteAddr();
	
	
	
	/**
	 * Returns the query string that is contained in the request URL 
	 * after the path. This method returns <code>null</code> 
	 * if the URL does not have a query string.
	 * 
	 * @return a String containing the query string or null if the URL contains no query string. 
	 */
	String getQueryString();
	
	
	
	/**
	 * Returns the part of this request's URL from the protocol
	 * name up to the query string in the first line of the HTTP request.
	 * 
	 * @return a String containing the part of the URL from the protocol 
	 *         name up to the query string
	 */
	String getRequestURI();

	
	
	/**
	 * Reconstructs the URL the client used to make the request.
	 * 
	 * @return the URL
	 */
	URL getRequestUrl();
	
	
	/**
	 * set the request url
	 * 
	 * @param url the request url
	 */
	void setRequestUrl(URL url);
	
	
	
	/**
	 * Returns a boolean indicating whether this request 
	 * was made using a secure channel, such as HTTPS.
	 *   
	 * @return a boolean indicating if the request was made using a secure channel
	 */
	boolean isSecure();
	
	
	
	/**
	 * Returns the name of the HTTP method with
	 * which this request was made, for example, GET, POST, or PUT.
	 * 
	 * @return a String specifying the name of the method 
	 */
	String getMethod();
	
	
	
	/**
	 * Sets the name of the HTTP method 
	 * 
	 * @param method a String specifying the name of the method
	 */
	void setMethod(String method);

	
	/**
	 * returns the Host header parameter or <code>null</code> if the header is not set
     *
     * @return the Host header parameter or <code>null</code> if the header is not set
	 */
	String getHost();

	
	/**
	 * sets the Host header
	 * 
	 * @param host the Host header
	 */
	void setHost(String host);
	
	
	/**
	 * returns the User-Agent header parameter or <code>null</code> if the header is not set
     *
     * @return the User-Agent header parameter or <code>null</code> if the header is not set
	 */
	String getUserAgent();
	
	
	/**
	 * sets the User-Agent header
	 * @param userAgent the User-Agent header
	 */
	void setUserAgent(String userAgent);
	
	
	/**
	 * Returns the host name of the server to which the request was sent. 
	 * It is the value of the part before ":" in the Host  header value, 
	 * if any, or the resolved server name, or the server IP address.
	 * 
	 * @return the server name
	 */
	String getServerName();
	
	
	/**
	 * Returns the port number to which the request was sent. It is the 
	 * value of the part after ":" in the Host  header value, if any, 
	 * or the server port where the client connection was accepted on.
	 * 
	 * @return the server port
	 */
	int getServerPort();	
		
	
	
	/**
	 * sets a parameter 
	 * 
	 * @param parameterName   the parameter name 
	 * @param parameterValue  the parameter value
	 */
	void setParameter(String parameterName, String parameterValue);


	
	/**
	 * Returns an Enumeration of String  objects containing the names 
	 * of the parameters contained in this request. If the request has no 
	 * parameters, the method returns an empty Enumeration. 
	 * 
	 * @return an Enumeration of String  objects, each String containing 
	 *         the name of a request parameter; or an empty Enumeration if the request has no parameters
	 */
	@SuppressWarnings("unchecked")
	Enumeration getParameterNames();
	
	
	/**
	 * returns the parameter name set
	 *  
	 * @return the parameter name set
	 */
	Set<String> getParameterNameSet();
	
	
	
	/**
	 * Returns the value of a request parameter as a String, or null 
	 * if the parameter does not exist. Request parameters are extra 
	 * information sent with the request. 
	 * 
	 * @param name a String specifying the name of the parameter
	 * @return a String representing the single value of the parameter 
	 */
	String getParameter(String name);
	
	

	/**
	 * Returns an array of String objects containing all of the values 
	 * the given request parameter has, or null if the parameter does not exist.
	 * If the parameter has a single value, the array has a length of 1.
	 *  
	 * @param name a String specifying the name of the parameter
	 * @return an array of String objects containing the parameter's values
	 */
	String[] getParameterValues(String name);
	
	
	
	/**
     * Get an string parameter or throws an exception if parameter is not present
	 * 
	 * @param name    the parameter name
	 * @return the value
	 * @throws BadMessageException if the parameter is not present
	 */
	String getRequiredStringParameter(String name) throws BadMessageException;

	
	/**
	 * Get an Integer parameter, or null if not present.  
	 * 
	 * @param name the name of the parameter
	 * @return the value, or <code>null</code>
	 * @throws BadMessageException  if the parameter value is not a number
	 */
	Integer getIntParameter(String name) throws BadMessageException;
	
	
	
	/**
     * Get an int parameter or throws an exception if parameter is not present
	 * 
	 * @param name    the parameter name
	 * @return the value
	 * @throws BadMessageException if the parameter is not present or the parameter is not a number
	 */
	int getRequiredIntParameter(String name) throws BadMessageException;
	
	
	/**
	 * Get an int parameter, with a fallback value.  
	 * 
	 * @param name        the name of the parameter
	 * @param defaultVal  the default value to use as fallback
	 * @return the value, or <code>null</code>
	 */
	int getIntParameter(String name, int defaultVal);
	
	
	
	/**
	 * Get an Long parameter, or null if not present.  
	 * 
	 * @param name the name of the parameter
	 * @return the value, or <code>null</code>
	 * @throws BadMessageException  if the parameter value is not a number 
	 */
	Long getLongParameter(String name) throws BadMessageException;
	
	
	/**
     * Get an long parameter or throws an exception if parameter is not present
	 * 
	 * @param name    the parameter name
	 * @return the value
	 * @throws BadMessageException if the parameter is not present or the parameter is not a number
	 */
	long getRequiredLongParameter(String name) throws BadMessageException;

	
	
	/**
	 * Get an long parameter, with a fallback value.  
	 * 
	 * @param name        the name of the parameter
	 * @param defaultVal  the default value to use as fallback
	 * @return the value, or <code>null</code>
	 */
	long getLongParameter(String name, long defaultVal);
	
	

	
	/**
	 * Get an Double parameter, or null if not present.  
	 * 
	 * @param name the name of the parameter
	 * @return the value, or <code>null</code>
	 * @throws BadMessageException if the parameter is not a number 
	 */
	Double getDoubleParameter(String name) throws BadMessageException;
	
	
	/**
     * Get an double parameter or throws an exception if parameter is not present
	 * 
	 * @param name    the parameter name
	 * @return the value
	 * @throws BadMessageException if the parameter is not present or the parameter value is not a number
	 */
	double getRequiredDoubleParameter(String name) throws BadMessageException;

	
	/**
	 * Get an double parameter, with a fallback value.  
	 * 
	 * @param name        the name of the parameter
	 * @param defaultVal  the default value to use as fallback
	 * @return the value, or <code>null</code>
	 */
	double getDoubleParameter(String name, double defaultVal);
	
	
	
	
	/**
	 * Get an Float parameter, or null if not present.  
	 * 
	 * @param name the name of the parameter
	 * @return the value, or <code>null</code>
	 * @throws BadMessageException  if the parameter value is not a number 
	 */
	Float getFloatParameter(String name) throws BadMessageException;
	
	
	/**
     * Get an float parameter or throws an exception if parameter is not present
	 * 
	 * @param name    the parameter name
	 * @return the value
	 * @throws BadMessageException if the parameter is not present or the parameter value is not a number
	 */
	float getRequiredFloatParameter(String name) throws BadMessageException;

	
	
	/**
	 * Get an float parameter, with a fallback value.  
	 * 
	 * @param name        the name of the parameter
	 * @param defaultVal  the default value to use as fallback
	 * @return the value, or <code>null</code>
	 */
	float getFloatParameter(String name, float defaultVal);
	

	
	
	/**
	 * Get an Boolean parameter, or null if not present.  
	 * 
	 * @param name the name of the parameter
	 * @return the value, or <code>null</code>
	 */
	Boolean getBooleanParameter(String name);
	
	
	/**
     * Get an boolean parameter or throws an exception if parameter is not present
	 * 
	 * @param name    the parameter name
	 * @return the value
	 * @throws BadMessageException if the parameter is not present
	 */
	boolean getRequiredBooleanParameter(String name) throws BadMessageException;

	
	
	
	/**
	 * Get an boolean parameter, with a fallback value.  
	 * 
	 * @param name        the name of the parameter
	 * @param defaultVal  the default value to use as fallback
	 * @return the value, or <code>null</code>
	 */
	boolean getBooleanParameter(String name, boolean defaultVal);

	
	/**
	 * creates a copy of this header 
	 * 
	 * @return return the copy
	 */
	IHttpRequestHeader copy();
}

 