/*
 *  Copyright (c) xlightweb.org, 2008. All rights reserved.
 *
 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Lesser General Public
 *  License as published by the Free Software Foundation; either
 *  version 2.1 of the License, or (at your option) any later version.
 *
 *  This library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Lesser General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public
 *  License along with this library; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Please refer to the LGPL license at: http://www.gnu.org/copyleft/lesser.txt
 * The latest copy of this software may be found on http://www.xlightweb.org/
 */
package org.xlightweb;

import java.io.File;
import java.io.IOException;
import java.io.RandomAccessFile;
import java.net.MalformedURLException;
import java.nio.ByteBuffer;
import java.nio.channels.FileChannel;




/**
 * POST request
 * 
 * @author grro@xlightweb.org
 */
public class PostRequest extends HttpRequest  {
	
	
	
	
	/**
	 * constructor 
	 * 
	 * @param url   the url string
	 * @throws MalformedURLException if the url is malformed
	 */
	public PostRequest(String url) throws MalformedURLException {
		super(new HttpRequestHeader("POST", url));
	}
	
	/**
	 * constructor 
	 * 
	 * @param url           the url string
	 * @param contentType   the content type
	 * @throws MalformedURLException if the url is malformed
	 * @throws IOException if an exception occurs
	 */
	public PostRequest(String url, String contentType) throws IOException, MalformedURLException {
		super("POST", url);
		setContentType(contentType);
	}

	
	/**
	 * constructor 
	 * 
	 * @param url           the url string
	 * @param contentType   the content type
	 * @param body          the body
	 * @throws MalformedURLException if the url is malformed
	 * @throws IOException if an exception occurs  
	 */
	public PostRequest(String url, String contentType, String body) throws IOException, MalformedURLException {
		super("POST", url, contentType, body);		
	}
	
	
	/**
	 * constructor 
	 * 
	 * @param url           the url string
	 * @param file          the file 
	 * @throws MalformedURLException if the url is malformed
	 * @throws IOException if an exception occurs 
	 */
	public PostRequest(String url, File file) throws IOException, MalformedURLException {
		super(new HttpRequestHeader("POST", url), new NonBlockingBodyDataSource("ISO-8859-1"));

		setContentType(getContentTypeByFileExtension(file));
		setContentLength((int) file.length());
		removeHeader("Transfer-Encoding");
		
		RandomAccessFile raf = new RandomAccessFile(file, "r");
		FileChannel fc = raf.getChannel();

		ByteBuffer buffer = ByteBuffer.allocate(getContentLength());
		fc.read(buffer);
		fc.close();
		raf.close();
		buffer.flip();
		
		getNonBlockingBody().append(buffer);
		getNonBlockingBody().setComplete(true);
	}
	
	
	
	/**
	 * constructor 
	 * 
	 * @param url           the url string
	 * @param encoding      the encoding
	 * @param contentType   the content type
	 * @param body          the body 
	 * @throws MalformedURLException if the url is malformed
	 * @throws IOException if an exception occurs 
	 */
	public PostRequest(String url, String contentType, String encoding, String body) throws IOException, MalformedURLException {
		super("POST", url, contentType, encoding, body);		
	}
	
	
	/**
	 * constructor 
	 * 
	 * @param url           the url string 
	 * @param contentType   the content type
	 * @param body          the body 
	 * @throws MalformedURLException if the url is malformed
	 * @throws IOException if an exception occurs 
	 */
	public PostRequest(String url, String contentType, byte[] body) throws IOException, MalformedURLException {
		super("POST", url, contentType, body);
	}
	
	

	
	/**
	 * constructor 
	 * 
	 * @param url           the url string
	 * @param contentType   the content type
	 * @param body          the body
	 * @throws MalformedURLException if the url is malformed
	 * @throws IOException if an exception occurs  
	 */
	public PostRequest(String url, String contentType, ByteBuffer[] body) throws IOException, MalformedURLException {
		super("POST", url, contentType, body);
	}
	
	
	/**
	 * constructor<br><br>
	 * 
	 * Example:
	 * <pre>
	 * PostRequest req = new PostRequest(urlBase + "/login", new String[] { "name=ItsMe", "pwd=VerySecret" }));
	 * ...
	 * </pre>
	 * 
	 *    
	 * @param url             the url string 
	 * @param formParameters  array of FORM parameters formatted as <key>=<value>
	 * @throws MalformedURLException if the url is malformed
	 * @throws IOException if an exception occurs 
	 */
	public PostRequest(String url, String[] formParameters) throws IOException, MalformedURLException {
		super("POST", url, "application/x-www-form-urlencoded; charset=utf-8", createFormUrlEncodedBody(formParameters));
	}
}
